
#include <stdio.h>
#include <ncurses.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <pthread.h>
#include <time.h>
#include <dlfcn.h>

#include "config.h"
#include "ncgui.h"
#include "pfregex.h"
#include "pfqueue.h"
#include "pfqmessage.h"
#include "backends/pfq_backend.h"

void *beptr;
char* (*pfbe_id)();
char* (*pfbe_version)();
int (*pfbe_init)(int,char**);
int (*pfbe_close)();
int (*pfbe_setup)(int,int,struct msg_t*,struct msg_t*);
int (*pfbe_fill_queue)();
int (*pfbe_retr_headers)(int);
int (*pfbe_retr_body)(int,char*,size_t);
int (*pfbe_message_delete)(int);
int (*pfbe_message_hold)(int);
int (*pfbe_message_release)(int);
int (*pfbe_message_requeue)(int);
int (*pfbe_set_queue)(int);
char* (*pfbe_queue_get_name)(int);
void (*pfbe_use_envelope)(int);
int (*pfbe_get_caps)();
int (*pfbe_queue_count)();

int dig_suspend;
int page_step;

struct msg_t *queue;
struct msg_t *queue_thread;
char *cat_buf;

char backends_path[100];

void version() {
	printf ( "pfqueue version %s\n(C) 2004,2005 Stefano Rivoir\n", VERSION );
}

void usage() {
	version();
	printf ( "Usage: pfqueue [-ehv] [-b postfix1|postfix2|exim] [-q queue_num] [-m max_msg]\n"
		 "       [-s autorefresh_seconds] [-l limit_seconds] -[B backends_path]\n"
		 "       [-p executables_pah] [-c config_path]\n" );
}

void help() {
	WINDOW *w;
	char *ht;
	int c;
	char *help_be;

	help_be = (char*)(malloc ( 1024 ) );
	sprintf ( help_be, "  Backend info:\n"
			"  ID: %s\n"
			"  Version: %s",
			pfbe_id(), pfbe_version() );

	w = newwin ( 25, 60, (LINES-25)/2,(COLS-60)/2 );

	nocbreak();
	cbreak();

	ht = help_text1;

	c = 0;
	while ( c!='q' ) {
		wclear ( w );
		mvwprintw ( w, 1, 0, ht );
		wnd_border ( w );
		wattron ( w, A_BOLD );
		mvwprintw ( w, 0, 28, "Help" );
		mvwprintw ( w, 0, 2, "pfqueue ver. " );
		mvwprintw ( w, 0, 15, VERSION );
		mvwprintw ( w, 23, 2, "Press q to exit, any key to next page" );
		wattroff ( w, A_BOLD );
		wrefresh ( w );
		c = wgetch ( w );
		if ( ht==help_text1 )
			ht = help_text2;
		else if ( ht == help_text2 )
			ht = help_be;
		else
			ht = help_text1;
	}
	delwin ( w );

	halfdelay ( delay_time*10 );
}

// Returns 1 if the file exists
int msg_fileex(int n) {
	return ( stat (queue[n].path, &foostat )!=-1);
}

void msg_cat() {
	WINDOW *w;
	char lines[MSG_MAX_ROWS][MSG_MAX_COLS];
	int j, i;
	int c;
	int fline;
	int pgstep;
	int msglines;
	int ptr;
	int buflen;
	
	if (!msg_fileex(CURMSG))
		return;

	memset ( lines, 0, sizeof(lines) );
	
	buflen = pfbe_retr_body ( CURMSG,cat_buf,1024*5 );
	msglines = 0;
	ptr = 0;

	// Copy cat_buf into splitted lines
	while ( ptr<buflen && msglines < MSG_MAX_ROWS ) {
		c = 0;
		while ( c < MSG_MAX_COLS && *(cat_buf+ptr)!='\n' ) {
			lines[msglines][c] = *(cat_buf+ptr);
			ptr++;
			c++;
		}
		while ( *(cat_buf+ptr)!='\n' )
			ptr++;
		ptr++;
		msglines++;
	}

	w = newwin ( LINES-4, COLS-4, 2, 2 );
	keypad ( w, TRUE );
	wnd_border (w);
	pgstep = LINES-4-2-2;
	
	wattron ( w, A_BOLD );
	mvwaddstr ( w, 0, 2, "Details for" );
	mvwaddstr ( w, 0, 14, queue[CURMSG].path );
	wattroff ( w, A_BOLD );

	c = 0;
	fline = 0;

	while ( c!='q' ) {
		j = fline;
		i = 2;
		while ( j<MSG_MAX_ROWS && i < LINES-4-1 ) {
			mvwaddnstr ( w, i, 2, lines[j++], COLS-4-3 );
			wclrtoeol ( w );
			mvwaddstr ( w, i++, COLS-4-1, "|" );
		}
		wrefresh ( w );

no_fill:
		c = wgetch (w);
		switch (c) {
		case 'g':
		case KEY_HOME:
			fline = 0;
			break;
		case 'G':
		case KEY_END:
			fline = msglines-pgstep;
			break;
		case KEY_PPAGE:
			if ( fline > pgstep )
				fline -= pgstep;
			else
				fline = 0;
			break;
		case KEY_NPAGE:
			if ( fline < (msglines-pgstep) )
				fline += pgstep;
			break;
		case KEY_UP:
			if ( fline )
				fline--;
			break;
		case KEY_DOWN:
			if ( fline<msglines-pgstep )
				fline++;
			break;
		case 'q':
			break;
		default:
			goto no_fill;
			break;
		}
	}
	delwin ( w );

	halfdelay ( delay_time*10 );
}

// Do thread dirty work
void* dir_dig_start(void *arg) {
	int NUMMSG_THREAD;
	int i;
	
	while ( 1 ) {
		if ( !dig_suspend && DO_SCAN ) {
			if ( dig_limit )
				dig_start = time(NULL);
			NUMMSG_THREAD = pfbe_fill_queue();
			NUMMSG = NUMMSG_THREAD;
			for ( i=0; i<NUMMSG_THREAD; i++ ) {
				if ( !queue_thread[i].cached )
					memcpy ( queue+i, queue_thread+i, sizeof(struct msg_t) );
			}
			dig_lastqueue = CURQ;
		}
		sleep ( delay_time );
	}
	pthread_mutex_unlock ( &dig_mutex );
	pthread_exit(NULL);
}

// Launches the dig thread
int fill_list() {
	if ( pthread_mutex_trylock(&dig_mutex)!=0 )
		return -1;
	pthread_create ( &dig_thread, NULL, dir_dig_start, NULL );
	return 0;
}

void goto_msg(int i) {
int j;
	PREVROW = CURROW;
	CURMSG  = i;
	CURROW  = (i-FIRSTMSG)+FIRSTROW;

	if ( CURROW<FIRSTROW ) {
		CURROW = FIRSTROW;
		FIRSTMSG = CURMSG;
	}
	if ( CURROW>MAXROW ) {
		j = CURROW-MAXROW;
		FIRSTMSG += j;
		CURROW = MAXROW;
	}
}

void goto_next() {
	if ( CURMSG<NUMMSG-1 )
		goto_msg( CURMSG+1 );
}

void goto_prev() {
	if ( CURMSG )
		goto_msg( CURMSG-1 );
}

void goto_first() {
	goto_msg(0);
}

void goto_last() {
	if ( NUMMSG )
		goto_msg( NUMMSG-1 );
}
void goto_nextpage() {
	if ( CURMSG+page_step < NUMMSG-1 )
		goto_msg ( CURMSG+page_step );
	else
		goto_msg ( NUMMSG-1 );
}
void goto_prevpage() {
	if ( CURMSG > page_step )
		goto_msg ( CURMSG-page_step );
	else
		goto_msg ( 0 );
}

// Performs an action on the message
void msg_action_do ( int i, int act ) {

	if ( !ISROOT )
		return;
	if ( !NUMMSG )
		return;

	switch ( act ) {
	case MSG_DELETE:
		pfbe_message_delete ( i );
		break;
	case MSG_HOLD:
		pfbe_message_hold ( i );
		break;
	case MSG_RELEASE:
		pfbe_message_release ( i );
		break;
	case MSG_REQUEUE:
		pfbe_message_requeue ( i );
		break;
	default:
		return;
	}
}

// Tags all messages
void msg_tagall() {
int i;
	for ( i=0; i<NUMMSG; i++ ) {
		pfbe_retr_headers ( i );
		queue[i].tagged = TRUE;
	}
	NUMTAG = NUMMSG;
}

// Resets tag flag on all messages
void msg_tagreset() {
int i;
	for ( i=0; i<NUMMSG; i++ )
		queue[i].tagged = FALSE;
	WRK_TAGGED = FALSE;
	NUMTAG = 0;
}

// Reset the cache flag of the messages
void msg_cachereset() {
int i;
	for ( i=0; i<NUMMSG; i++ )
		queue[i].cached = 0;
}

// Wrapper
void msg_action ( int act ) {
int i;

	if ( !ISROOT )
		return;

	if ( !WRK_TAGGED && !msg_fileex(CURMSG) )
		return;

	if ( ASK_CONFIRM && !wnd_confirm ( a_names[act] ) )
		return;

	if ( WRK_TAGGED ) {
		dig_suspend = 1;
		for ( i = 0; i<NUMMSG; i++ ) {
			if ( queue[i].tagged )
				msg_action_do ( i, act );
		}
		msg_tagreset();
		dig_suspend = 0;
		CURROW = FIRSTROW;
		CURMSG = 0;
	} else
		msg_action_do ( CURMSG, act );
	return;
}


// Sets the list window title
void win_listtitle() {
	char buf[BUF_SIZE];
	wattron ( lwnd, A_BOLD );

	sprintf ( buf, "Queue: '%s', %d message%s, %d tagged %s %s %s",
			pfbe_queue_get_name(CURQ), NUMMSG, 
			( NUMMSG != 1 ? "s" : "" ),
			NUMTAG,
			( WRK_TAGGED ? "*Commands work on tagged*" : "" ),
			( ASK_CONFIRM ? "" : "*Confirmation disabled*" ),
			( DO_SCAN ? "" : "*Scan suspended*" ) );
	mvwprintw ( lwnd, 0, 1, buf );

	wclrtoeol ( lwnd );
	wattron ( lwnd, A_BOLD );
	mvwaddstr ( lwnd, 1, 1, "ID" );
	mvwaddstr ( lwnd, 1, COL1, "From" );
	mvwaddstr ( lwnd, 1, COL2, "To" );
	wclrtoeol ( lwnd );
	wattroff ( lwnd, A_BOLD );
	wnoutrefresh ( lwnd );
}

// Sizes the window according to the (new) screen size
void win_resize () {

	int h;

	h = 8;
	
	MAXROW = LINES-h-1;

	if ( mwndb )
		delwin ( mwndb );
	if ( lwnd )
		delwin ( lwnd );
	if ( mwnd )
		delwin ( mwnd );
	if ( bline )
		free ( bline );

	mainwnd = initscr();

	bline = (char*)malloc(sizeof(char)*COLS);
	memset ( bline, ' ', sizeof(char)*COLS-1 );
	*(bline+COLS-1) = 0;

	mwndb = newwin ( h, COLS, LINES-h, 0 );
	lwnd  = newwin ( LINES-h, COLS, 0, 0 );
	mwnd  = newwin ( h-2, COLS-2, LINES-h+1, 1 );

	if ( use_envelope )
		wnd_clear ( mwndb, "Message (from envelope)" );
	else
		wnd_clear ( mwndb, "Message (from headers)" );

	COL1 = 20;
	COL2 = COL1+( COLS - COL1 ) / 2-2;
	
	wnoutrefresh ( mwndb );

	page_step = MAXROW -2;

	doupdate();
}

// Clears the cursor in the previous position and displays the current one
void show_cursor() {
	mvwaddstr ( lwnd, PREVROW, 0, " " );
	mvwaddstr ( lwnd, CURROW,  0, ">" );
}

// Changes the current queue
void queue_change( int q ) {

	if ( q > pfbe_queue_count() )
		return;
	
	CURROW = PREVROW = FIRSTROW;
	FIRSTMSG = 0;
	CURMSG = 0;
	CURQ = q;
	NUMTAG = 0;
	WRK_TAGGED = FALSE;
	
	pfbe_set_queue ( q );

	// Ensure that a scan is made before proceeding, loop 1/5 sec
	while ( dig_lastqueue != CURQ ) { usleep ( 200000 ); };
	
	win_resize();
	msg_tagreset();
	win_listtitle();

}

// Displays the bottom window
void msg_show () {

	if ( !NUMMSG ) {
		wclear ( mwnd );
		mvwaddstr ( mwnd, 0, 0, "Queue is empty" );
		return;
	}
	
	mvwaddstr ( mwnd, 0, 0, "Message:" );
	mvwaddstr ( mwnd, 0, 9, queue[CURMSG].id );
	wclrtoeol ( mwnd );

	mvwaddstr ( mwnd, 1, 0, "From..:" );
	mvwaddstr ( mwnd, 1, 8, queue[CURMSG].from );
	wclrtoeol ( mwnd );

	mvwaddstr ( mwnd, 2, 0, "To....:" );
	mvwaddstr ( mwnd, 2, 8, queue[CURMSG].to );
	wclrtoeol ( mwnd );

	mvwaddstr ( mwnd, 3, 0, "Subj..:" );
	mvwaddstr ( mwnd, 3, 8, queue[CURMSG].subj );
	wclrtoeol ( mwnd );

	mvwaddstr ( mwnd, 4, 0, "Status:" );
	mvwaddstr ( mwnd, 5, 0, " " );
	wclrtoeol ( mwnd );
	mvwaddstr ( mwnd, 4, 8, queue[CURMSG].stat );
	wclrtoeol ( mwnd );
	
}

// Displays the list of messages
void queue_show() {
int i, j;
time_t show_start;

	j = FIRSTROW;
	i = FIRSTMSG;

	show_start = time(NULL);
	while ( j <= MAXROW && ((dig_limit && time(NULL)-show_start<=dig_limit) || !dig_limit) ) {

		if ( i<NUMMSG ) {
			pfbe_retr_headers ( i );
			if ( queue[i].tagged )
				wattron ( lwnd, A_BOLD );
			if ( i == CURMSG || i == MSGMARK )
				wattron ( lwnd, A_REVERSE );
			mvwaddstr  ( lwnd, j, 0, bline );
			mvwaddnstr ( lwnd, j, 1, queue[i].id, sizeof(queue[i].id) );
			mvwaddnstr ( lwnd, j, COL1,   queue[i].from, COL2-2-COL1 );
			mvwaddstr  ( lwnd, j, COL2-2, "  " );
			mvwaddnstr ( lwnd, j, COL2,   queue[i].to, COLS-2-COL2 );
			wattroff ( lwnd, A_BOLD | A_REVERSE );
		} else {
			mvwaddstr ( lwnd, j, 1, " " );
			wclrtoeol ( lwnd );
		}
		j++;
		i++;
	}

	if ( NUMMSG && CURMSG > (NUMMSG-1) ) {
		goto_last();
	}
	keypad ( lwnd, 1 );
	win_listtitle();
	show_cursor();
	msg_show();
	wnoutrefresh ( mwnd );
	wnoutrefresh ( lwnd );
	doupdate();
}

// Search for a message; returns -1 if not found
int msg_match(int s, int direction) {
int i, res;

	res = 0;

	if ( direction==0 ) {
		for ( i=s; i<NUMMSG; i++ ) {
			pfbe_retr_headers ( i );
			if ( ( (search_mode & SM_FROM) && regexec(regexp,queue[i].from,res,NULL,0)==0 ) ||
			     ( (search_mode & SM_TO)   && regexec(regexp,queue[i].to  ,res,NULL,0)==0 ) ||
			     ( (search_mode & SM_SUBJ) && regexec(regexp,queue[i].subj,res,NULL,0)==0 ) ) {
				goto_msg ( i );
				queue_show();
				return i;
			}
		}
	} else {
		for ( i=s; i>=0; i-- ) {
			pfbe_retr_headers ( i );
			if ( ( (search_mode & SM_FROM) && regexec(regexp,queue[i].from,res,NULL,0)==0 ) ||
			     ( (search_mode & SM_TO)   && regexec(regexp,queue[i].to  ,res,NULL,0)==0 ) ||
			     ( (search_mode & SM_SUBJ) && regexec(regexp,queue[i].subj,res,NULL,0)==0 ) ) {
				goto_msg ( i );
				queue_show();
				return i;
			}
		}
	}
	return -1;
}

void msg_search() {
int res;

	wnd_getregexp();
	res = regcomp ( regexp, regexps, 0 );
	if ( !res ) 
		msg_match(CURMSG,0);
}

int msg_searchnext() {
	if ( !strlen(regexps) )
		return -1;
	return msg_match(CURMSG+1,0);
}

int msg_searchprev() {
	if ( !strlen(regexps) )
		return -1;
	return msg_match(CURMSG-1,1);
}

void msg_searchandtag() {
int res;

	wnd_getregexp();
	res = regcomp ( regexp, regexps, 0 );
	if ( !res ) {
		res = msg_match(0,0);
		while ( res != -1 ) {
			queue[res].tagged = 1;
			NUMTAG++;
			res = msg_searchnext(res,0);
		}
	}
}

// Toggle envelope usage for from/to
void toggle_envelope() {
	if ( !(pfbe_get_caps() & BECAPS_MSG_ENVELOPE) )
		return;

	use_envelope = !use_envelope;
	msg_cachereset();
	pfbe_use_envelope ( use_envelope );
	win_resize();
	queue_show();
}

// Main loop
void main_loop() {
int c;
int i;

	c = 0;
	
	CURMSG = 0;
	FIRSTMSG = 0;

	wnd_clear ( mwndb, "Message" );

	while ( c != 'q' ) {

		queue_show();
		c = wgetch ( lwnd );
		PREVROW = CURROW;
		
		switch ( c ) {
		case 'G':
		case KEY_END:
			goto_last();
			break;
		case 'g':
		case KEY_HOME:
			goto_first();
			break;
		case KEY_UP:
			goto_prev();
			break;
		case KEY_DOWN:
			goto_next();
			break;
		case KEY_PPAGE:
			goto_prevpage();
			break;
		case KEY_NPAGE:
			goto_nextpage();
			break;
		case KEY_RESIZE:
			win_resize();
			break;
		case 'e':
			toggle_envelope();
			break;
		case 'd':
			msg_action( MSG_DELETE );
			break;
		case 'h':
			msg_action( MSG_HOLD );
			break;
		case 'l':
			msg_action ( MSG_RELEASE );
			break;
		case 'r':
			msg_action ( MSG_REQUEUE );
			break;
		case '1':
			queue_change ( Q_DEFERRED );
			break;
		case '4':
			queue_change ( Q_HOLD );
			break;
		case '3':
			queue_change ( Q_INCOMING );
			break;
		case '2':
			queue_change ( Q_ACTIVE );
			break;
		case 'a':
			msg_tagall();
			break;
		case '.':
			MSGMARK = CURMSG;
			queue[CURMSG].tagged = TRUE;
			queue_show();
			break;
		case 't':
			if ( MSGMARK==-1 ) {
				if ( queue[CURMSG].tagged )
					NUMTAG--;
				else
					NUMTAG++;
				queue[CURMSG].tagged = !queue[CURMSG].tagged;
			} else {
				if ( CURMSG>=MSGMARK ) {
					for ( i=MSGMARK; i<=CURMSG; i++ ) {
						queue[i].tagged = TRUE;
					}
				} else {
					for ( i=CURMSG; i<=MSGMARK; i++ ) {
						queue[i].tagged = TRUE;
					}
				}
			}
			MSGMARK = -1;
			queue_show();
			break;
		case 'T':
			msg_searchandtag();
			break;
		case 'u':
			msg_tagreset();
			queue_show();
			break;
		case ';':
			WRK_TAGGED = !WRK_TAGGED;
			queue_show();
			break;
		case '/':
			msg_search();
			break;
		case 'n':
			msg_searchnext();
			break;
		case 'p':
			msg_searchprev();
			break;
		case '?':
			help();
			break;
		case 'c':
			ASK_CONFIRM = !ASK_CONFIRM;
			break;
		case '-':
			DO_SCAN = !DO_SCAN;
			break;
		case 'q':
		case 'Q':
			break;
		case 's':
		case 'S':
			if ( NUMMSG ) {
				msg_cat();
				win_resize();
				msg_show();
			}
			break;
		}
		
	}
}

int load_backend ( const char* be ) {
	char buf[BUF_SIZE];
	
	if ( strlen(backends_path) )
		sprintf ( buf, "%s/libpfq_%s.so", backends_path, be );
	else
		sprintf ( buf, "%s/libpfq_%s.so", PFBEDIR, be );
	
	beptr = dlopen ( buf, RTLD_LAZY );
	if ( !beptr ) {
		fprintf ( stderr, dlerror() );
		fprintf ( stderr, "\n" );
		// Try 'pfqueue' subdir
		if ( strlen(backends_path) )
			sprintf ( buf, "%s/pfqueue/libpfq_%s.so", backends_path, be );
		else
			sprintf ( buf, "pfqueue/libpfq_%s.so", be );
		beptr = dlopen ( buf, RTLD_LAZY );
	}
	if ( !beptr ) {
		fprintf ( stderr, dlerror() );
		fprintf ( stderr, "\n" );
		return -2;
	}
	
	pfbe_init  = dlsym( beptr, "pfb_init" );
	if ( !pfbe_init )
		return -1;

	pfbe_id  = dlsym( beptr, "pfb_id" );
	if ( !pfbe_id )
		return -1;

	pfbe_version  = dlsym( beptr, "pfb_version" );
	if ( !pfbe_version )
		return -1;

	pfbe_setup = dlsym( beptr, "pfb_setup" );
	if ( !pfbe_setup )
		return -1;

	pfbe_fill_queue = dlsym( beptr, "pfb_fill_queue" );
	if ( !pfbe_fill_queue )
		return -1;

	pfbe_retr_headers = dlsym( beptr, "pfb_retr_headers" );
	if ( !pfbe_retr_headers )
		return -1;

	pfbe_retr_body = dlsym( beptr, "pfb_retr_body" );
	if ( !pfbe_retr_body )
		return -1;

	pfbe_message_delete = dlsym( beptr, "pfb_message_delete" );
	if ( !pfbe_message_delete )
		return -1;

	pfbe_message_hold = dlsym( beptr, "pfb_message_hold" );
	if ( !pfbe_message_hold )
		return -1;

	pfbe_message_release = dlsym( beptr, "pfb_message_release" );
	if ( !pfbe_message_release )
		return -1;

	pfbe_message_requeue = dlsym( beptr, "pfb_message_requeue" );
	if ( !pfbe_message_requeue )
		return -1;

	pfbe_set_queue = dlsym( beptr, "pfb_set_queue" );
	if ( !pfbe_set_queue )
		return -1;

	pfbe_use_envelope = dlsym( beptr, "pfb_use_envelope" );
	if ( !pfbe_use_envelope )
		return -1;

	pfbe_get_caps = dlsym( beptr, "pfb_get_caps" );
	if ( !pfbe_get_caps )
		return -1;
	
	pfbe_queue_get_name = dlsym( beptr, "pfb_queue_get_name" );
	if ( !pfbe_queue_get_name )
		return -1;
	
	pfbe_queue_count = dlsym( beptr, "pfb_queue_count" );
	if ( !pfbe_queue_count )
		return -1;

	return 0;

}

int try_backend ( const char *b, int argc, char** argv ) {
	int res;

	res = load_backend ( b );
	if ( res )
		return 0;
	res = pfbe_init ( argc, argv );
	if ( res )
		return 0;
	return 1;
}

int main ( int argc, char** argv ) {
	int opt;
	int hd, initial_queue;
	char be[BUF_SIZE];
	int res;

	// Some defaults
	MSGMARK = -1;
	ASK_CONFIRM = TRUE;
	MSG_MAX = 200;
	FIRSTROW = 2;
	hd = 2;
	use_envelope = FALSE;
	ISROOT = (getuid()==0);
	initial_queue = 0;
	dig_suspend = 0;
	dig_limit = 0;
	DO_SCAN = 1;

	// Init mutex
	pthread_mutex_init ( &dig_mutex, NULL );

	regexps = (char*) malloc ( REGEXP_LEN );
	regexp  = (regex_t*) malloc (sizeof(regex_t));

	strcpy ( be, "autodetect" );
	memset ( backends_path, 0, sizeof(backends_path) );

	opt = 0;
	while ( opt!=-1 ) {
		opt = getopt ( argc, argv, "B:b:q:s:m:l:ehv" );
		switch (opt) {
			case 'B':
				snprintf ( backends_path, sizeof(backends_path), "%s", optarg );
				break;
			case 'b':
				snprintf ( be, sizeof(be), "%s", optarg );
				break;
			case 'q':
				initial_queue = atoi ( optarg );
				switch ( initial_queue ) {
				case 1:
					initial_queue = 0;
					break;
				case 2:
					initial_queue = 3;
					break;
				case 3:
					initial_queue = 2;
					break;
				case 4:
					initial_queue = 1;
					break;
				default:
					initial_queue = 0;
				}
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'v':
				version();
				exit(0);
				break;
			case 'm':
				MSG_MAX = atoi ( optarg );
				if ( MSG_MAX > 2048 )
					MSG_MAX = 2048;
				if ( MSG_MAX < 5 )
					MSG_MAX = 5;
				break;
			case 's':
				hd = atoi ( optarg );
				if ( hd > 300 )
					hd = 300;
				if ( hd < 1 )
					hd = 1;
				break;
			case 'l':
				dig_limit = atoi ( optarg );
				if ( dig_limit > 300 )
					dig_limit = 300;
				if ( dig_limit < 1 )
					dig_limit = 1;
				break;
			case 'e':
				use_envelope = TRUE;
				break;
		}
	}

	NUMTAG = 0;
	WRK_TAGGED = FALSE;
	FIRSTMSG = 0;
	mainwnd = NULL;
	mwndb = lwnd = mwnd = NULL;
	bline = NULL;

	queue = (struct msg_t*)malloc(sizeof(struct msg_t)*MSG_MAX);
	if ( !queue ) {
		fprintf ( stderr, "pfqueue: sorry, cannot malloc for %d elements!\n", MSG_MAX );
		exit (-2);
	}
	queue_thread = (struct msg_t*)malloc(sizeof(struct msg_t)*MSG_MAX);
	if ( !queue_thread ) {
		free ( queue );
		fprintf ( stderr, "pfqueue: sorry, cannot malloc for %d elements!\n", MSG_MAX );
		exit (-2);
	}
	cat_buf = (char*)malloc(1012*5);
	if ( !cat_buf ) {
		free ( queue );
		free ( queue_thread );
		fprintf ( stderr, "pfqueue, sorry cannot malloc for cat buffer\n" );
		exit (-2);
	}

	if ( !strcmp ( be, "autodetect" ) ) {
		// Try to autodetect backend
		strcpy ( be, "exim" );
		res = try_backend ( be, argc, argv );
		if ( !res ) {
			strcpy ( be, "postfix2" );
			res = try_backend ( be, argc, argv );
		}
		if ( !res ) {
			strcpy ( be, "postfix1" );
			res = try_backend ( be, argc, argv );
		}
		if ( !res ) {
			fprintf ( stderr, "pfqueue: cannot autodetect suitable backend, try -b and/or -B option\n" );
			exit (1);
		}
	}
	
	switch ( load_backend ( be ) ) {
	case -2:
		fprintf ( stderr, "pfqueue: backend not found!\n" );
		exit (1);
		break;
	case -1:
		fprintf ( stderr, "pfqueue: backend not valid (missing symbols)!\n" );
		exit (1);
		break;
	};
	

	if ( pfbe_init(argc,argv) ) {
		fprintf ( stderr, "pfqueue: %s backend failed to init!\n", be );
		exit (-3);
	}
	
	pfbe_setup( MSG_MAX, dig_limit, queue, queue_thread );

	win_resize();
	halfdelay ( hd*10 );
	delay_time = hd;
	noecho();
	
	dig_lastqueue = -1;
	fill_list();
	queue_change ( initial_queue );

	fill_list();
	main_loop();

	pthread_mutex_destroy ( &dig_mutex );
	if (beptr)
		dlclose (beptr);

	free ( cat_buf );
	free ( queue );
	free ( queue_thread );
	free ( bline );
	free ( regexps );
	regfree ( regexp );

	endwin();

	return 0;

}
